from urllib.parse import urlencode

try:
    import autopep8
except ImportError:
    autopep8 = None
from django.template import Context, Template

from silk.profiling.dynamic import is_str_typ

template = """\
from django.test import Client
c = Client()
response = c.{{ lower_case_method }}(path='{{ path }}'{% if data or content_type %},{% else %}){% endif %}{% if data %}
data={{ data }}{% endif %}{% if data and content_type %},{% elif data %}){% endif %}{% if content_type %}
content_type='{{ content_type }}'){% endif %}
"""


def _encode_query_params(query_params):
    try:
        query_params = urlencode(query_params)
    except TypeError:
        pass
    return '?' + query_params


def gen(path, method=None, query_params=None, data=None, content_type=None):
    # generates python code representing a call via django client.
    # useful for use in testing
    method = method.lower()
    t = Template(template)
    context = {
        'path': path,
        'lower_case_method': method,
        'content_type': content_type,
    }
    if method == 'get':
        context['data'] = query_params
    else:
        if query_params:
            query_params = _encode_query_params(query_params)
            path += query_params
        if is_str_typ(data):
            data = "'%s'" % data
        context['data'] = data
        context['query_params'] = query_params
    code = t.render(Context(context, autoescape=False))
    if autopep8:
        # autopep8 is not a hard requirement, so we check if it's available
        # if autopep8 is available, we use it to format the code and do things
        # like remove long lines and improve readability
        code = autopep8.fix_code(
            code,
            options=autopep8.parse_args(['--aggressive', '']),
        )
    return code



import re

_RE_STRINGS    = re.compile(r"'[^']*'|\"[^\"]*\"")
_RE_NUMBERS    = re.compile(r'\b\d+\b')
_RE_WHITESPACE = re.compile(r'\s+')


def fingerprint_query(sql: str) -> str:
    s = _RE_STRINGS.sub('?', sql)
    s = _RE_NUMBERS.sub('?', s)
    return _RE_WHITESPACE.sub(' ', s).strip().lower()


class NPlusOneGroup:
    __slots__ = ('fingerprint', 'queries', 'count', 'representative')

    def __init__(self, fingerprint, queries):
        self.fingerprint = fingerprint
        self.queries = queries
        self.count = len(queries)
        self.representative = queries[0]


class NPlusOneResult:
    __slots__ = ('groups', 'flagged_query_ids', 'has_n_plus_one')

    def __init__(self, groups):
        self.groups = groups
        self.flagged_query_ids = {q.pk for g in groups for q in g.queries}
        self.has_n_plus_one = bool(groups)


def detect_n_plus_one(queries, threshold=3) -> NPlusOneResult:
    buckets = {}
    for q in queries:
        buckets.setdefault(fingerprint_query(q.query), []).append(q)
    flagged = [NPlusOneGroup(fp, qs) for fp, qs in buckets.items() if len(qs) >= threshold]
    flagged.sort(key=lambda g: g.count, reverse=True)
    return NPlusOneResult(flagged)



from unittest.mock import Mock, NonCallableMagicMock, NonCallableMock, patch

from django.test import TestCase
from django.utils.encoding import force_str

from silk.collector import DataCollector
from silk.models import Request, SQLQuery
from silk.sql import execute_sql

from .util import delete_all_models

_simple_mock_query_sql = 'SELECT * FROM table_name WHERE column1 = %s'
_simple_mock_query_params = ('asdf',)
_non_unicode_binary_mock_query_params = (b'\x0a\x00\x00\xff',)
_unicode_binary_mock_query_params = ('🫠'.encode(),)


def mock_sql(mock_query_params):
    mock_sql_query = Mock(spec_set=['_execute_sql', 'query', 'as_sql', 'connection'])
    mock_sql_query._execute_sql = Mock()
    mock_sql_query.query = NonCallableMock(spec_set=['model'])
    mock_sql_query.query.model = Mock()
    mock_sql_query.as_sql = Mock(return_value=(_simple_mock_query_sql, mock_query_params))

    mock_sql_query.connection = NonCallableMock(
        spec_set=['cursor', 'features', 'ops'],
        cursor=Mock(
            spec_set=['__call__'],
            return_value=NonCallableMagicMock(spec_set=['__enter__', '__exit__', 'execute'])
        ),
        features=NonCallableMock(
            spec_set=['supports_explaining_query_execution'],
            supports_explaining_query_execution=True
        ),
        ops=NonCallableMock(spec_set=['explain_query_prefix'],
                            explain_query_prefix=Mock(return_value='')),
    )

    return mock_sql_query, mock_query_params


class BaseTestCase(TestCase):
    def tearDown(self):
        DataCollector().stop_python_profiler()

    def call_execute_sql(self, request, mock_query_params):
        DataCollector().configure(request=request)
        delete_all_models(SQLQuery)
        self.query_string = _simple_mock_query_sql
        self.mock_sql, self.query_params = mock_sql(mock_query_params)
        self.kwargs = {
            'one': 1,
            'two': 2
        }
        self.args = [1, 2]
        execute_sql(self.mock_sql, *self.args, **self.kwargs)


class TestCallNoRequest(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.call_execute_sql(None, _simple_mock_query_params)

    def test_called(self):
        self.mock_sql._execute_sql.assert_called_once_with(*self.args, **self.kwargs)

    def test_count(self):
        self.assertEqual(0, len(DataCollector().queries))


class TestCallRequest(BaseTestCase):
    def test_query_simple(self):
        self.call_execute_sql(Request(), _simple_mock_query_params)
        self.mock_sql._execute_sql.assert_called_once_with(*self.args, **self.kwargs)
        self.assertEqual(1, len(DataCollector().queries))
        query = list(DataCollector().queries.values())[0]
        expected = self.query_string % tuple(force_str(param) for param in self.query_params)
        self.assertEqual(query['query'], expected)

    def test_query_unicode(self):
        self.call_execute_sql(Request(), _unicode_binary_mock_query_params)
        self.mock_sql._execute_sql.assert_called_once_with(*self.args, **self.kwargs)
        self.assertEqual(1, len(DataCollector().queries))
        query = list(DataCollector().queries.values())[0]
        expected = self.query_string % tuple(force_str(param) for param in self.query_params)
        self.assertEqual(query['query'], expected)

    def test_query_non_unicode(self):
        self.call_execute_sql(Request(), _non_unicode_binary_mock_query_params)
        self.mock_sql._execute_sql.assert_called_once_with(*self.args, **self.kwargs)
        self.assertEqual(0, len(DataCollector().queries))


class TestCallSilky(BaseTestCase):
    def test_no_effect(self):
        DataCollector().configure()
        sql, _ = mock_sql(_simple_mock_query_params)
        sql.query.model = NonCallableMagicMock(spec_set=['__module__'])
        sql.query.model.__module__ = 'silk.models'
        # No SQLQuery models should be created for silk requests for obvious reasons
        with patch('silk.sql.DataCollector', return_value=Mock()) as mock_DataCollector:
            execute_sql(sql)
            self.assertFalse(mock_DataCollector().register_query.call_count)


class TestCollectorInteraction(BaseTestCase):
    def _query(self):
        try:
            query = list(DataCollector().queries.values())[0]
        except IndexError:
            self.fail('No queries created')
        return query

    def test_request(self):
        DataCollector().configure(request=Request.objects.create(path='/path/to/somewhere'))
        sql, _ = mock_sql(_simple_mock_query_params)
        execute_sql(sql)
        query = self._query()
        self.assertEqual(query['request'], DataCollector().request)

    def test_registration(self):
        DataCollector().configure(request=Request.objects.create(path='/path/to/somewhere'))
        sql, _ = mock_sql(_simple_mock_query_params)
        execute_sql(sql)
        query = self._query()
        self.assertIn(query, DataCollector().queries.values())

    def test_explain_simple(self):
        DataCollector().configure(request=Request.objects.create(path='/path/to/somewhere'))
        sql, params = mock_sql(_simple_mock_query_params)
        prefix = "EXPLAIN"
        mock_cursor = sql.connection.cursor.return_value.__enter__.return_value
        sql.connection.ops.explain_query_prefix.return_value = prefix
        execute_sql(sql)
        self.assertNotIn(prefix, params)
        mock_cursor.execute.assert_called_once_with(f"{prefix} {_simple_mock_query_sql}", params)

    def test_explain_unicode(self):
        DataCollector().configure(request=Request.objects.create(path='/path/to/somewhere'))
        sql, params = mock_sql(_unicode_binary_mock_query_params)
        prefix = "EXPLAIN"
        mock_cursor = sql.connection.cursor.return_value.__enter__.return_value
        sql.connection.ops.explain_query_prefix.return_value = prefix
        execute_sql(sql)
        self.assertNotIn(prefix, params)
        mock_cursor.execute.assert_called_once_with(f"{prefix} {_simple_mock_query_sql}", params)

    def test_explain_non_unicode(self):
        DataCollector().configure(request=Request.objects.create(path='/path/to/somewhere'))
        sql, params = mock_sql(_non_unicode_binary_mock_query_params)
        prefix = "EXPLAIN"
        mock_cursor = sql.connection.cursor.return_value.__enter__.return_value
        sql.connection.ops.explain_query_prefix.return_value = prefix
        execute_sql(sql)
        self.assertNotIn(prefix, params)
        self.assertFalse(mock_cursor.execute.called)



import json

from django.db.models import Sum
from django.shortcuts import render
from django.template.context_processors import csrf
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.models import Request, Response, SQLQuery
from silk.request_filters import (
    BaseFilter,
    FiltersManager,
    TIME_RANGE_PRESETS,
    filters_from_request,
)
from silk.utils.n_plus_one import fingerprint_query
from silk.utils.pagination import get_page

__author__ = 'mtford'

# Sort options: maps field key → display label + optional queryset transform
SORT_OPTIONS = {
    'start_time': {
        'label': 'Recent',
        'additional_query_filter': None,
    },
    'path': {
        'label': 'Path',
        'additional_query_filter': None,
    },
    'num_sql_queries': {
        'label': 'Num. Queries',
        'additional_query_filter': None,
    },
    'time_taken': {
        'label': 'Time',
        'additional_query_filter': lambda qs: qs.filter(time_taken__gte=0),
    },
    'db_time': {
        'label': 'DB Time',
        'additional_query_filter': lambda qs: qs.annotate(
            db_time=Sum('queries__time_taken')
        ).filter(db_time__gte=0),
    },
}

DEFAULT_SORT_LIST = [{'field': 'start_time', 'dir': 'DESC'}]
DEFAULT_PER_PAGE = 25
PER_PAGE_OPTIONS = [10, 25, 50, 100]
SESSION_KEY_FILTERS = 'request_filters'
SESSION_KEY_SORT = 'request_sort'


def _apply_sort(query_set, sort_list):
    """Apply a list of {field, dir} dicts as ORDER BY to *query_set*.

    Annotations are applied per-field (they're additive), but all ORDER BY
    columns are passed in a single .order_by() call so they compound rather
    than replace each other.
    """
    order_fields = []
    for item in sort_list:
        field = item.get('field', 'start_time')
        direction = item.get('dir', 'DESC')
        if field not in SORT_OPTIONS:
            continue
        opt = SORT_OPTIONS[field]
        if opt['additional_query_filter'] is not None:
            query_set = opt['additional_query_filter'](query_set)
        prefix = '-' if direction == 'DESC' else ''
        order_fields.append(f'{prefix}{field}')
    if order_fields:
        query_set = query_set.order_by(*order_fields)
    return query_set


class RequestsView(View):
    # Keep legacy attributes for backward compatibility
    show = PER_PAGE_OPTIONS
    default_show = DEFAULT_PER_PAGE

    order_by = {
        k: {'label': v['label'], 'additional_query_filter': v['additional_query_filter']}
        for k, v in SORT_OPTIONS.items()
    }
    order_dir = {
        'ASC': {'label': 'Ascending'},
        'DESC': {'label': 'Descending'},
    }
    view_style = {
        'card': {'label': 'Cards'},
        'row': {'label': 'Rows'},
    }
    default_order_by = 'start_time'
    default_order_dir = 'DESC'
    default_view_style = 'row'

    session_key_request_filters = SESSION_KEY_FILTERS
    filters_manager = FiltersManager(SESSION_KEY_FILTERS)

    @property
    def options_order_by(self):
        return [{'value': x, 'label': SORT_OPTIONS[x]['label']} for x in SORT_OPTIONS]

    @property
    def options_order_dir(self):
        return [{'value': x, 'label': self.order_dir[x]['label']} for x in self.order_dir]

    @property
    def options_view_style(self):
        return [{'value': x, 'label': self.view_style[x]['label']} for x in self.view_style]

    def _get_paths(self):
        return Request.objects.values_list('path', flat=True).order_by('path').distinct()

    def _get_views(self):
        return (
            Request.objects.values_list('view_name', flat=True)
            .exclude(view_name='')
            .order_by('view_name')
            .distinct()
        )

    def _get_status_codes(self):
        return (
            Response.objects.values_list('status_code', flat=True)
            .order_by('status_code')
            .distinct()
        )

    def _get_methods(self):
        return Request.objects.values_list('method', flat=True).order_by('method').distinct()

    def _get_sort_list(self, request):
        """Return the current sort list from session, falling back to defaults."""
        if hasattr(request, 'session'):
            return request.session.get(SESSION_KEY_SORT, DEFAULT_SORT_LIST)
        return DEFAULT_SORT_LIST

    def _save_sort_list(self, request, sort_list):
        if hasattr(request, 'session'):
            request.session[SESSION_KEY_SORT] = sort_list

    # Sentinel so we can distinguish "not passed" from explicit None
    _SHOW_NOT_SET = object()

    def _get_objects(self, show=_SHOW_NOT_SET, order_by=None, order_dir=None, path=None, filters=None, sort_list=None, request=None):
        """Build and return a queryset (or sliced queryset for backward compat).

        When *show* is not passed, defaults to self.default_show and slices the
        queryset — preserving the original API so existing tests still pass.
        Pass show=None explicitly to get the full (unsliced) queryset for the paginator.
        """
        if not filters:
            filters = []
        query_set = Request.objects.all()

        # Determine sort
        if sort_list:
            query_set = _apply_sort(query_set, sort_list)
        else:
            if not order_by:
                order_by = self.default_order_by
            if not order_dir:
                order_dir = self.default_order_dir
            if order_by not in SORT_OPTIONS:
                raise RuntimeError('Unknown order_by: "%s"' % order_by)
            query_set = _apply_sort(query_set, [{'field': order_by, 'dir': order_dir}])

        if path:
            query_set = query_set.filter(path=path)
        for f in filters:
            query_set = f.contribute_to_query_set(query_set)
            query_set = query_set.filter(f)

        # Backward-compatible slicing: if show was not explicitly passed, default to default_show
        if show is self._SHOW_NOT_SET:
            show = self.default_show
        if show is not None:
            return query_set[:show]
        return query_set

    def _create_context(self, request):
        raw_filters = self.filters_manager.get(request).copy()
        show = raw_filters.pop('show', self.default_show)
        order_by = raw_filters.pop('order_by', self.default_order_by)
        order_dir = raw_filters.pop('order_dir', self.default_order_dir)
        view_style = raw_filters.pop('view_style', self.default_view_style)

        if show:
            show = int(show)
        per_page = show or DEFAULT_PER_PAGE

        path = request.GET.get('path', None)
        sort_list = self._get_sort_list(request)

        qs = self._get_objects(
            show=None,  # paginator handles limiting
            order_by=order_by,
            order_dir=order_dir,
            path=path,
            filters=[BaseFilter.from_dict(x) for _, x in raw_filters.items()],
            sort_list=sort_list,
        )

        page_obj = get_page(request, qs, per_page)

        # Batch N+1 detection for the current page (1 extra DB query)
        page_request_ids = [r.pk for r in page_obj.object_list]
        has_n1_on_page = False
        if page_request_ids:
            sql_rows = SQLQuery.objects.filter(
                request_id__in=page_request_ids
            ).values('request_id', 'query')
            # Group by request, fingerprint, count occurrences
            buckets = {}  # request_id -> fingerprint -> count
            for row in sql_rows:
                rid = row['request_id']
                fp = fingerprint_query(row['query'])
                buckets.setdefault(rid, {}).setdefault(fp, 0)
                buckets[rid][fp] += 1
            n1_request_ids = {
                rid for rid, fps in buckets.items()
                if any(cnt >= 3 for cnt in fps.values())
            }
            for r in page_obj.object_list:
                r.has_n1 = r.pk in n1_request_ids
            has_n1_on_page = bool(n1_request_ids)

        context = {
            'show': show,
            'per_page': per_page,
            'order_by': order_by,
            'order_dir': order_dir,
            'view_style': view_style,
            'request': request,
            'options_show': self.show,
            'options_order_by': self.options_order_by,
            'options_order_dir': self.options_order_dir,
            'options_view_style': self.options_view_style,
            'options_paths': self._get_paths(),
            'options_status_codes': self._get_status_codes(),
            'options_methods': self._get_methods(),
            'view_names': self._get_views(),
            'filters': raw_filters,
            'sort_list': sort_list,
            'sort_fields_used': [item['field'] for item in sort_list],
            'sort_options': [
                {'value': k, 'label': v['label']} for k, v in SORT_OPTIONS.items()
            ],
            'time_presets': [
                {'key': k, 'seconds': v, 'label': k} for k, v in TIME_RANGE_PRESETS.items()
            ],
            'page_obj': page_obj,
            'results': page_obj.object_list,  # backward compat
            'has_n1_on_page': has_n1_on_page,
        }
        context.update(csrf(request))
        if path:
            context['path'] = path
        return context

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request):
        # Read sort_criteria JSON from URL param (enables shared URLs to restore sort state)
        if request.GET.get('sort_criteria'):
            try:
                sort_list = json.loads(request.GET['sort_criteria'])
                if isinstance(sort_list, list) and sort_list:
                    self._save_sort_list(request, sort_list)
            except (ValueError, TypeError):
                pass
        # Handle sort_criteria from GET (backward compat: order_by / order_dir)
        elif request.GET.get('order_by') or request.GET.get('order_dir'):
            ob = request.GET.get('order_by', self.default_order_by)
            od = request.GET.get('order_dir', self.default_order_dir)
            self._save_sort_list(request, [{'field': ob, 'dir': od}])

        if request.GET:
            filters = {
                **self.filters_manager.get(request),
                **{
                    k: v
                    for k, v in request.GET.items()
                    if k in ['show', 'order_by', 'order_dir', 'view_style']
                },
            }
            # per_page in URL is an alias for show (enables shared URLs to restore page size)
            if 'per_page' in request.GET and 'show' not in request.GET:
                filters['show'] = request.GET['per_page']
            self.filters_manager.save(request, filters)
        return render(request, 'silk/requests.html', self._create_context(request))

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def post(self, request):
        previous_session = self.filters_manager.get(request)

        # Handle sort_criteria update from POST
        sort_criteria_raw = request.POST.get('sort_criteria', '')
        if sort_criteria_raw:
            try:
                sort_list = json.loads(sort_criteria_raw)
                if isinstance(sort_list, list):
                    self._save_sort_list(
                        request, sort_list if sort_list else DEFAULT_SORT_LIST
                    )
            except (ValueError, TypeError):
                pass

        # Detect whether this POST came from the filter form or the sort/toolbar form.
        # The filter form always includes at least one filter-*-typ hidden field.
        # The sort/toolbar form only contains sort_criteria, show, view_style.
        is_filter_submit = any(k.startswith('filter-') for k in request.POST)

        if 'clear_filters' in request.POST:
            filters = {k: v for k, v in previous_session.items() if k in ['show', 'view_style']}
            self.filters_manager.save(request, filters)
        elif is_filter_submit:
            # Full filter form submitted — replace filter set with what the form says.
            filters = {
                **{k: v for k, v in previous_session.items() if k in ['show', 'order_by', 'order_dir', 'view_style']},
                **{ident: f.as_dict() for ident, f in filters_from_request(request).items()},
            }
            self.filters_manager.save(request, filters)
        else:
            # Sort / toolbar form — only update show/view_style, keep all existing filters.
            non_filter_keys = ['show', 'order_by', 'order_dir', 'view_style']
            updated = {k: v for k, v in request.POST.items() if k in non_filter_keys}
            filters = {**previous_session, **updated}
            self.filters_manager.save(request, filters)

        return render(request, 'silk/requests.html', self._create_context(request))



import json

from django.db.models import Avg, Count, Max, Min, Sum
from django.db.models.functions import TruncHour, TruncDay
from django.shortcuts import render
from django.template.context_processors import csrf
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk import models
from silk.auth import login_possibly_required, permissions_possibly_required
from silk.request_filters import (
    BaseFilter,
    FiltersManager,
    TIME_RANGE_PRESETS,
    filters_from_request,
)


def _percentile(sorted_data, p):
    """Linear-interpolation percentile (stdlib only, works on SQLite/PG/MySQL)."""
    n = len(sorted_data)
    if n == 0:
        return 0.0
    idx = (p / 100.0) * (n - 1)
    lo = int(idx)
    hi = lo + 1
    frac = idx - lo
    if hi >= n:
        return float(sorted_data[-1])
    return sorted_data[lo] * (1 - frac) + sorted_data[hi] * frac


class SummaryView(View):
    filters_key = 'summary_filters'
    filters_manager = FiltersManager(filters_key)

    def _avg_num_queries(self, filters):
        queries__aggregate = models.Request.objects.filter(*filters).annotate(num_queries=Count('queries')).aggregate(num=Avg('num_queries'))
        return queries__aggregate['num']

    def _avg_time_spent_on_queries(self, filters):
        taken__aggregate = models.Request.objects.filter(*filters).annotate(time_spent=Sum('queries__time_taken')).aggregate(num=Avg('time_spent'))
        return taken__aggregate['num']

    def _avg_overall_time(self, filters):
        taken__aggregate = models.Request.objects.filter(*filters).annotate(time_spent=Sum('time_taken')).aggregate(num=Avg('time_spent'))
        return taken__aggregate['num']

    # TODO: Find a more efficient way to do this. Currently has to go to DB num. views + 1 times and is prob quite expensive
    def _longest_query_by_view(self, filters):
        values_list = models.Request.objects.filter(*filters).values_list("view_name").annotate(max=Max('time_taken')).filter(max__isnull=False).order_by('-max')[:5]
        requests = []
        for view_name, _ in values_list:
            request = models.Request.objects.filter(view_name=view_name, *filters).filter(time_taken__isnull=False).order_by('-time_taken')[0]
            requests.append(request)
        return sorted(requests, key=lambda item: item.time_taken, reverse=True)

    def _time_spent_in_db_by_view(self, filters):
        values_list = models.Request.objects.filter(*filters).values_list('view_name').annotate(t=Sum('queries__time_taken')).filter(t__gte=0).order_by('-t')[:5]
        requests = []
        for view, _ in values_list:
            r = models.Request.objects.filter(view_name=view, *filters).annotate(t=Sum('queries__time_taken')).filter(t__isnull=False).order_by('-t')[0]
            requests.append(r)
        return sorted(requests, key=lambda item: item.t, reverse=True)

    def _request_time_percentiles(self, filters):
        times = list(
            models.Request.objects.filter(*filters)
            .filter(time_taken__isnull=False)
            .order_by('time_taken')
            .values_list('time_taken', flat=True)
        )
        return {p: round(_percentile(times, p), 2) for p in [25, 50, 75, 95, 99]}

    def _sql_time_percentiles(self, filters):
        pks = list(models.Request.objects.filter(*filters).values_list('pk', flat=True))
        if not pks:
            return {p: 0.0 for p in [25, 50, 75, 95, 99]}
        times = list(
            models.SQLQuery.objects.filter(request_id__in=pks, time_taken__isnull=False)
            .order_by('time_taken')
            .values_list('time_taken', flat=True)
        )
        return {p: round(_percentile(times, p), 2) for p in [25, 50, 75, 95, 99]}

    def _request_timeline(self, filters):
        """Hourly (or daily) request counts for the activity chart."""
        qs = models.Request.objects.filter(*filters)
        agg = qs.aggregate(min_t=Min('start_time'), max_t=Max('start_time'))
        if not agg['min_t']:
            return []
        span_hours = (agg['max_t'] - agg['min_t']).total_seconds() / 3600
        trunc = TruncDay if span_hours > 72 else TruncHour
        buckets = (
            qs.annotate(bucket=trunc('start_time'))
            .values('bucket')
            .annotate(count=Count('id'))
            .order_by('bucket')
        )
        return [{'t': b['bucket'].isoformat(), 'count': b['count']} for b in buckets]

    def _status_distribution(self, filters):
        """Count requests by HTTP status class."""
        request_pks = list(models.Request.objects.filter(*filters).values_list('pk', flat=True))
        if not request_pks:
            return {'2xx': 0, '3xx': 0, '4xx': 0, '5xx': 0}
        rows = (
            models.Response.objects
            .filter(request_id__in=request_pks, status_code__isnull=False)
            .values('status_code')
            .annotate(count=Count('id'))
        )
        buckets = {'2xx': 0, '3xx': 0, '4xx': 0, '5xx': 0}
        for row in rows:
            sc = row['status_code']
            if 200 <= sc < 300:
                buckets['2xx'] += row['count']
            elif 300 <= sc < 400:
                buckets['3xx'] += row['count']
            elif 400 <= sc < 500:
                buckets['4xx'] += row['count']
            elif sc >= 500:
                buckets['5xx'] += row['count']
        return buckets

    def _method_distribution(self, filters):
        """Count requests by HTTP method, sorted descending."""
        rows = (
            models.Request.objects.filter(*filters)
            .values('method')
            .annotate(count=Count('id'))
            .order_by('-count')
        )
        return [{'method': r['method'], 'count': r['count']} for r in rows]

    def _response_time_histogram(self, filters):
        """Bucket response times into 6 fixed ranges."""
        times = list(
            models.Request.objects.filter(*filters)
            .filter(time_taken__isnull=False, time_taken__gte=0)
            .values_list('time_taken', flat=True)
        )
        buckets = [
            {'label': '<50ms',    'max': 50,   'count': 0},
            {'label': '50-100ms', 'max': 100,  'count': 0},
            {'label': '100-200ms','max': 200,  'count': 0},
            {'label': '200-500ms','max': 500,  'count': 0},
            {'label': '500ms-1s', 'max': 1000, 'count': 0},
            {'label': '>1s',      'max': None,  'count': 0},
        ]
        prev = 0
        for t in times:
            for b in buckets:
                if b['max'] is None or t < b['max']:
                    b['count'] += 1
                    break
        return [{'label': b['label'], 'count': b['count']} for b in buckets]

    def _num_queries_by_view(self, filters):
        queryset = models.Request.objects.filter(*filters).values_list('view_name').annotate(t=Count('queries')).order_by('-t')[:5]
        views = [r[0] for r in queryset[:6]]
        requests = []
        for view in views:
            try:
                r = models.Request.objects.filter(view_name=view, *filters).annotate(t=Count('queries')).order_by('-t')[0]
                requests.append(r)
            except IndexError:
                pass
        return sorted(requests, key=lambda item: item.t, reverse=True)

    def _create_context(self, request):
        raw_filters = self.filters_manager.get(request)
        filters = [BaseFilter.from_dict(filter_d) for _, filter_d in raw_filters.items()]
        avg_overall_time = self._avg_num_queries(filters)
        num_requests = models.Request.objects.filter(*filters).count()
        active_preset = raw_filters.get('time_preset', {}).get('value')
        request_percentiles = self._request_time_percentiles(filters)
        sql_percentiles = self._sql_time_percentiles(filters)
        c = {
            'request': request,
            'num_requests': num_requests,
            'num_profiles': models.Profile.objects.filter(*filters).count(),
            'avg_num_queries': avg_overall_time,
            'avg_time_spent_on_queries': self._avg_time_spent_on_queries(filters),
            'avg_overall_time': self._avg_overall_time(filters),
            'longest_queries_by_view': self._longest_query_by_view(filters),
            'most_time_spent_in_db': self._time_spent_in_db_by_view(filters),
            'most_queries': self._num_queries_by_view(filters),
            'filters': raw_filters,
            'has_data': num_requests > 0,
            'time_presets': [
                {'key': k, 'seconds': v, 'label': k} for k, v in TIME_RANGE_PRESETS.items()
            ],
            'active_preset': active_preset,
            'request_percentiles': request_percentiles,
            'sql_percentiles': sql_percentiles,
            'chart_json': json.dumps({
                'request': request_percentiles,
                'sql': sql_percentiles,
                'timeline': self._request_timeline(filters),
                'status': self._status_distribution(filters),
                'methods': self._method_distribution(filters),
                'rt_hist': self._response_time_histogram(filters),
            }),
        }
        c.update(csrf(request))
        return c

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request):
        c = self._create_context(request)
        return render(request, 'silk/summary.html', c)

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def post(self, request):
        # Handle time preset shortcut
        preset_key = request.POST.get('time_preset', '')
        if preset_key in TIME_RANGE_PRESETS:
            from silk.request_filters import SecondsFilter
            seconds = TIME_RANGE_PRESETS[preset_key]
            f = SecondsFilter(seconds)
            filters = {'time_preset': f.as_dict()}
        elif 'clear_filters' in request.POST:
            filters = {}
        else:
            filters = {ident: f.as_dict() for ident, f in filters_from_request(request).items()}
        self.filters_manager.save(request, filters)
        return render(request, 'silk/summary.html', self._create_context(request))



from django.core.management.base import BaseCommand

import silk.models
from silk.config import SilkyConfig


class Command(BaseCommand):
    help = "Triggers silk's request garbage collect."

    def add_arguments(self, parser):
        parser.add_argument(
            "-m",
            "--max-requests",
            default=SilkyConfig().SILKY_MAX_RECORDED_REQUESTS,
            type=int,
            help="Maximum number of requests to keep after garbage collection.",
        )

    def handle(self, *args, **options):
        if "max_requests" in options:
            max_requests = options["max_requests"]
            SilkyConfig().SILKY_MAX_RECORDED_REQUESTS = max_requests
        if options["verbosity"] >= 2:
            max_requests = SilkyConfig().SILKY_MAX_RECORDED_REQUESTS
            request_count = silk.models.Request.objects.count()
            self.stdout.write(
                f"Keeping up to {max_requests} of {request_count} requests."
            )
        silk.models.Request.garbage_collect(force=True)



from django.urls import path

from . import views

app_name = 'example_app'
urlpatterns = [
    path(route='', view=views.index, name='index'),
    path(route='create', view=views.ExampleCreateView.as_view(), name='create'),
]



#!/usr/bin/env python3
"""
migrate_to_silky.py
-------------------
Automated zero-data-loss migration from django-silk to django-silky.

Supports PostgreSQL, MySQL / MariaDB, and SQLite.

Usage
-----
  # From your Django project root (where manage.py lives):
  python migrate_to_silky.py

  # Specify manage.py location explicitly:
  python migrate_to_silky.py --manage-py /path/to/manage.py

  # Custom backup directory:
  python migrate_to_silky.py --backup-dir /var/backups/silk

  # Install a specific version:
  python migrate_to_silky.py --silky-version 1.0.0

  # See what would happen without touching anything:
  python migrate_to_silky.py --dry-run

  # Skip backup (not recommended, but useful in CI with external backups):
  python migrate_to_silky.py --skip-backup
"""

import argparse
import json
import os
import shutil
import subprocess
import sys
from datetime import datetime
from pathlib import Path

# ── Constants ─────────────────────────────────────────────────────────────────

OLD_PACKAGE = "django-silk"
NEW_PACKAGE = "django-silky"

# All tables owned by the 'silk' Django app
SILK_TABLES = [
    "silk_request",
    "silk_response",
    "silk_sqlquery",
    "silk_profile",
]

# Expected migration names (0001–0008) — same in both packages
EXPECTED_MIGRATIONS = [
    "0001_initial",
    "0002_auto_update_uuid4_id_field",
    "0003_request_prof_file",
    "0004_request_prof_file_storage",
    "0005_increase_request_prof_file_length",
    "0006_fix_request_poc_file_blank",
    "0007_sqlquery_identifier",
    "0008_sqlquery_analysis",
]

# ── Terminal colours ──────────────────────────────────────────────────────────

_USE_COLOUR = sys.stdout.isatty() and os.environ.get("NO_COLOR") is None

def _c(code: str) -> str:
    return code if _USE_COLOUR else ""

RESET  = _c("\033[0m")
BOLD   = _c("\033[1m")
GREEN  = _c("\033[32m")
YELLOW = _c("\033[33m")
RED    = _c("\033[31m")
CYAN   = _c("\033[36m")
DIM    = _c("\033[2m")


def info(msg: str)    -> None: print(f"{CYAN}  ▶  {RESET}{msg}")
def ok(msg: str)      -> None: print(f"{GREEN}  ✔  {RESET}{msg}")
def warn(msg: str)    -> None: print(f"{YELLOW}  ⚠  {RESET}{msg}")
def err(msg: str)     -> None: print(f"{RED}  ✖  {RESET}{msg}", file=sys.stderr)
def heading(msg: str) -> None: print(f"\n{BOLD}{msg}{RESET}")
def dim(msg: str)     -> None: print(f"{DIM}      {msg}{RESET}")


def die(msg: str, code: int = 1) -> None:
    err(msg)
    sys.exit(code)


# ── Subprocess helpers ────────────────────────────────────────────────────────

def run(
    cmd: list[str],
    capture: bool = False,
    check: bool = True,
    env: dict | None = None,
) -> subprocess.CompletedProcess:
    kwargs: dict = {"check": check, "env": env or os.environ.copy()}
    if capture:
        kwargs["capture_output"] = True
        kwargs["text"] = True
    return subprocess.run(cmd, **kwargs)


def pip_run(*args: str, capture: bool = False) -> subprocess.CompletedProcess:
    return run([sys.executable, "-m", "pip", *args], capture=capture)


def manage_run(
    manage_py: Path,
    *args: str,
    capture: bool = False,
    check: bool = True,
    env: dict | None = None,
) -> subprocess.CompletedProcess:
    return run([sys.executable, str(manage_py), *args], capture=capture, check=check, env=env)


# ── Locate manage.py ─────────────────────────────────────────────────────────

def find_manage_py(hint: str | None) -> Path:
    if hint:
        p = Path(hint)
        if not p.is_file():
            die(f"manage.py not found at: {hint}")
        return p.resolve()

    for candidate in [Path("manage.py"), *sorted(Path(".").glob("*/manage.py"))]:
        if candidate.is_file():
            return candidate.resolve()

    die(
        "Could not find manage.py.\n"
        "  Run this script from your Django project root, or pass "
        "--manage-py /path/to/manage.py"
    )


# ── Read Django DB config ─────────────────────────────────────────────────────

def get_db_config(manage_py: Path) -> dict:
    """Ask Django to serialise DATABASES['default'] as JSON."""
    snippet = (
        "import json; from django.conf import settings; "
        "db = dict(settings.DATABASES.get('default', {})); "
        "print('__SILKY_DB__' + json.dumps(db))"
    )
    result = manage_run(
        manage_py, "shell", "-c", snippet,
        capture=True, check=False,
    )
    if result.returncode != 0:
        die(
            "Could not read Django settings.\n"
            f"  stderr: {result.stderr.strip()}\n\n"
            "  Make sure:\n"
            "  • DJANGO_SETTINGS_MODULE is set (or use --settings)\n"
            "  • All dependencies are installed in the current Python environment"
        )
    for line in result.stdout.splitlines():
        line = line.strip()
        if line.startswith("__SILKY_DB__"):
            try:
                return json.loads(line[len("__SILKY_DB__"):])
            except json.JSONDecodeError:
                break
    die(
        "Unexpected output from Django shell.\n"
        f"  stdout: {result.stdout[:500]}"
    )


def detect_engine(db: dict) -> str:
    engine = db.get("ENGINE", "")
    if "postgresql" in engine or "psycopg" in engine:
        return "postgresql"
    if "mysql" in engine:
        return "mysql"
    if "sqlite" in engine:
        return "sqlite"
    die(
        f"Unsupported database engine: {engine!r}\n"
        "  Supported: django.db.backends.postgresql, .mysql, .sqlite3"
    )


# ── Backup helpers ────────────────────────────────────────────────────────────

def _pg_conn_args(db: dict) -> list[str]:
    """pg_dump / pg_restore connection flags (no db name)."""
    args: list[str] = []
    if db.get("HOST"):  args += ["--host",     db["HOST"]]
    if db.get("PORT"):  args += ["--port",     str(db["PORT"])]
    if db.get("USER"):  args += ["--username", db["USER"]]
    return args


def _pg_env(db: dict) -> dict:
    env = os.environ.copy()
    if db.get("PASSWORD"):
        env["PGPASSWORD"] = db["PASSWORD"]
    return env


def backup_postgresql(db: dict, backup_file: Path) -> None:
    table_flags: list[str] = []
    for t in SILK_TABLES:
        table_flags += ["--table", t]

    cmd = [
        "pg_dump",
        *_pg_conn_args(db),
        db["NAME"],
        "--format=custom",
        "--no-owner",
        "--no-acl",
        *table_flags,
        f"--file={backup_file}",
    ]
    dim("$ " + " ".join(cmd))
    run(cmd, env=_pg_env(db))
    ok(f"PostgreSQL backup → {backup_file}")


def restore_postgresql(db: dict, backup_file: Path) -> None:
    cmd = [
        "pg_restore",
        "--clean",
        "--if-exists",
        "--no-owner",
        "--no-acl",
        *_pg_conn_args(db),
        f"--dbname={db['NAME']}",
        str(backup_file),
    ]
    dim("$ " + " ".join(cmd))
    run(cmd, env=_pg_env(db))
    ok("PostgreSQL tables restored.")


def _mysql_conn_args(db: dict) -> list[str]:
    """mysql / mysqldump connection flags (password via env)."""
    args: list[str] = []
    if db.get("HOST"):  args += ["--host",  db["HOST"]]
    if db.get("PORT"):  args += ["--port",  str(db["PORT"])]
    if db.get("USER"):  args += ["--user",  db["USER"]]
    return args


def _mysql_env(db: dict) -> dict:
    env = os.environ.copy()
    if db.get("PASSWORD"):
        env["MYSQL_PWD"] = db["PASSWORD"]  # avoids password in ps output
    return env


def backup_mysql(db: dict, backup_file: Path) -> None:
    cmd = [
        "mysqldump",
        *_mysql_conn_args(db),
        "--single-transaction",
        "--add-drop-table",
        db["NAME"],
        *SILK_TABLES,
    ]
    dim("$ mysqldump [conn-args] " + db["NAME"] + " " + " ".join(SILK_TABLES))
    with open(backup_file, "w") as f:
        subprocess.run(cmd, stdout=f, check=True, env=_mysql_env(db))
    ok(f"MySQL backup → {backup_file}")


def restore_mysql(db: dict, backup_file: Path) -> None:
    cmd = ["mysql", *_mysql_conn_args(db), db["NAME"]]
    dim("$ mysql [conn-args] " + db["NAME"] + " < " + str(backup_file))
    with open(backup_file) as f:
        subprocess.run(cmd, stdin=f, check=True, env=_mysql_env(db))
    ok("MySQL tables restored.")


def backup_sqlite(db: dict, backup_file: Path) -> None:
    src = Path(db.get("NAME", ""))
    if not src.is_file():
        die(f"SQLite database not found: {src}")
    shutil.copy2(src, backup_file)
    ok(f"SQLite database copied → {backup_file}")


def restore_sqlite(db: dict, backup_file: Path) -> None:
    dest = Path(db.get("NAME", ""))
    shutil.copy2(backup_file, dest)
    ok(f"SQLite database restored → {dest}")


_BACKUP_FNS = {
    "postgresql": (backup_postgresql, restore_postgresql),
    "mysql":      (backup_mysql,      restore_mysql),
    "sqlite":     (backup_sqlite,     restore_sqlite),
}

_BACKUP_EXT = {
    "postgresql": ".dump",
    "mysql":      ".sql",
    "sqlite":     ".sqlite3",
}

_BACKUP_TOOL = {
    "postgresql": "pg_dump",
    "mysql":      "mysqldump",
    "sqlite":     None,  # no external tool needed
}


def do_backup(engine: str, db: dict, backup_dir: Path, dry_run: bool) -> Path:
    ts = datetime.now().strftime("%Y%m%d_%H%M%S")
    backup_file = backup_dir / f"silk_backup_{ts}{_BACKUP_EXT[engine]}"

    if dry_run:
        info(f"[dry-run] Would write backup → {backup_file}")
        return backup_file

    backup_dir.mkdir(parents=True, exist_ok=True)
    _BACKUP_FNS[engine][0](db, backup_file)
    return backup_file


def do_restore(engine: str, db: dict, backup_file: Path) -> None:
    warn(f"Restoring from backup: {backup_file}")
    _BACKUP_FNS[engine][1](db, backup_file)


# ── Package helpers ───────────────────────────────────────────────────────────

def get_version(package: str) -> str | None:
    """Return installed version of *package*, or None if not installed."""
    result = run(
        [
            sys.executable, "-c",
            f"import importlib.metadata; print(importlib.metadata.version({package!r}))",
        ],
        capture=True, check=False,
    )
    return result.stdout.strip() if result.returncode == 0 else None


# ── Migration helpers ─────────────────────────────────────────────────────────

def get_migration_status(manage_py: Path) -> dict[str, bool]:
    """Return {migration_label: applied} for the silk app."""
    result = manage_run(
        manage_py, "showmigrations", "silk", "--list",
        capture=True, check=False,
    )
    if result.returncode != 0:
        return {}

    status: dict[str, bool] = {}
    for line in result.stdout.splitlines():
        stripped = line.strip()
        if stripped.startswith("[X]"):
            status[stripped[4:].strip()] = True
        elif stripped.startswith("[ ]"):
            status[stripped[4:].strip()] = False
    return status


# ── Data verification ─────────────────────────────────────────────────────────

def verify_data(manage_py: Path) -> None:
    snippet = (
        "from silk.models import Request, SQLQuery, Profile; "
        "print(f'  requests : {Request.objects.count():,}'); "
        "print(f'  queries  : {SQLQuery.objects.count():,}'); "
        "print(f'  profiles : {Profile.objects.count():,}')"
    )
    result = manage_run(
        manage_py, "shell", "-c", snippet,
        capture=True, check=False,
    )
    if result.returncode == 0:
        for line in result.stdout.splitlines():
            if any(k in line for k in ("requests", "queries", "profiles")):
                ok(line.strip())
    else:
        warn("Could not query row counts — check /silk/ in the browser manually.")


# ── CLI ───────────────────────────────────────────────────────────────────────

def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="migrate_to_silky.py",
        description="Migrate from django-silk to django-silky with zero data loss.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="Full docs: https://github.com/VaishnavGhenge/django-silky/blob/master/MIGRATING.md",
    )
    parser.add_argument(
        "--manage-py", metavar="PATH",
        help="Path to manage.py (default: auto-detect)",
    )
    parser.add_argument(
        "--backup-dir", metavar="DIR", default="./silk_backups",
        help="Directory for backup files (default: ./silk_backups)",
    )
    parser.add_argument(
        "--silky-version", metavar="VER", default=None,
        help="django-silky version to install, e.g. 1.0.0 (default: latest)",
    )
    parser.add_argument(
        "--skip-backup", action="store_true",
        help="Skip the backup step (not recommended)",
    )
    parser.add_argument(
        "--dry-run", action="store_true",
        help="Show what would happen without making any changes",
    )
    return parser.parse_args()


# ── Main ──────────────────────────────────────────────────────────────────────

def main() -> None:
    args = parse_args()
    manage_py = find_manage_py(args.manage_py)
    backup_dir = Path(args.backup_dir)
    dry_run = args.dry_run
    install_spec = f"{NEW_PACKAGE}=={args.silky_version}" if args.silky_version else NEW_PACKAGE

    print(f"""
{BOLD}┌──────────────────────────────────────────────────┐
│  django-silk  →  django-silky  migration script  │
└──────────────────────────────────────────────────┘{RESET}""")

    if dry_run:
        warn("DRY-RUN mode — no changes will be made.\n")

    # ── Step 1: Pre-flight ───────────────────────────────────────────────────
    heading("Step 1 / 5 — Pre-flight checks")

    silk_ver = get_version(OLD_PACKAGE)
    silky_ver = get_version(NEW_PACKAGE)

    if silk_ver:
        ok(f"{OLD_PACKAGE} {silk_ver} installed")
    else:
        warn(f"{OLD_PACKAGE} not found — continuing anyway")

    if silky_ver:
        warn(f"{NEW_PACKAGE} {silky_ver} already installed — you may be re-running after a partial migration")

    info(f"manage.py: {manage_py}")

    db = get_db_config(manage_py)
    engine = detect_engine(db)
    ok(f"Database engine : {engine}  ({db.get('ENGINE', '')})")
    ok(f"Database name   : {db.get('NAME', '(unknown)')}")
    if db.get("HOST"):
        ok(f"Database host   : {db['HOST']}:{db.get('PORT', 'default')}")

    # Check backup tool is available
    if not args.skip_backup:
        tool = _BACKUP_TOOL[engine]
        if tool and not shutil.which(tool):
            die(
                f"Backup tool '{tool}' not found in PATH.\n\n"
                "  Install it and retry, or use --skip-backup to proceed without a backup.\n\n"
                "  PostgreSQL: apt install postgresql-client  |  brew install libpq\n"
                "  MySQL:      apt install default-mysql-client  |  brew install mysql-client"
            )
        if tool:
            ok(f"Backup tool     : {shutil.which(tool)}")
        else:
            ok("Backup tool     : built-in (shutil.copy2)")

    # ── Step 2: Backup ───────────────────────────────────────────────────────
    heading("Step 2 / 5 — Backup silk tables")

    backup_file: Path | None = None
    if args.skip_backup:
        warn("Skipping backup (--skip-backup passed).")
    else:
        backup_file = do_backup(engine, db, backup_dir, dry_run)

    # ── Step 3: Swap packages ────────────────────────────────────────────────
    heading("Step 3 / 5 — Swap packages")

    if dry_run:
        if silk_ver:
            info(f"[dry-run] pip uninstall {OLD_PACKAGE} -y")
        info(f"[dry-run] pip install {install_spec}")
    else:
        if silk_ver:
            info(f"Uninstalling {OLD_PACKAGE} {silk_ver} …")
            pip_run("uninstall", OLD_PACKAGE, "-y")
            ok(f"{OLD_PACKAGE} removed")
        else:
            info(f"{OLD_PACKAGE} not installed — skipping uninstall")

        info(f"Installing {install_spec} …")
        pip_run("install", install_spec)
        new_ver = get_version(NEW_PACKAGE) or "unknown"
        ok(f"{NEW_PACKAGE} {new_ver} installed")

    # ── Step 4: Migrations ───────────────────────────────────────────────────
    heading("Step 4 / 5 — Migrations")

    if dry_run:
        info("[dry-run] manage.py showmigrations silk")
        info("[dry-run] manage.py migrate silk  (only if pending)")
    else:
        migrations = get_migration_status(manage_py)

        if not migrations:
            warn(
                "Could not read migration status.\n"
                "  Run manually: python manage.py migrate silk"
            )
        else:
            applied = [m for m, done in migrations.items() if done]
            pending = [m for m, done in migrations.items() if not done]

            for m in applied:
                ok(f"[X] {m}")
            for m in pending:
                warn(f"[ ] {m}")

            if pending:
                info(f"Running: manage.py migrate silk  ({len(pending)} pending) …")
                result = manage_run(manage_py, "migrate", "silk", check=False)
                if result.returncode != 0:
                    err("manage.py migrate silk failed!")
                    if backup_file and backup_file.exists():
                        do_restore(engine, db, backup_file)
                        warn(f"Data restored. Re-install {OLD_PACKAGE} to get back to your previous state:")
                        warn(f"  pip install {OLD_PACKAGE}=={silk_ver or 'latest'}")
                    sys.exit(1)
                ok("All migrations applied.")
            else:
                ok(f"All {len(applied)} silk migrations already applied — no migrate needed.")

    # ── Step 5: Verify ───────────────────────────────────────────────────────
    heading("Step 5 / 5 — Verify data")

    if dry_run:
        info("[dry-run] Would query silk model row counts via manage.py shell")
    else:
        verify_data(manage_py)

    # ── Done ─────────────────────────────────────────────────────────────────
    rollback_cmd = f"pip install {OLD_PACKAGE}=={silk_ver}" if silk_ver else f"pip install {OLD_PACKAGE}"
    backup_note = str(backup_file) if backup_file else "(skipped)"

    print(f"""
{GREEN}{BOLD}  All done!{RESET}

  {BOLD}Next steps:{RESET}
  1. Restart your application server
  2. Visit /silk/ and confirm the UI loads with your historical data

  {BOLD}Backup location:{RESET}
  {backup_note}

  {BOLD}To roll back:{RESET}
  pip uninstall {NEW_PACKAGE} && {rollback_cmd}
  (no manage.py migrate needed — schema is shared)

  {DIM}Full migration docs: https://github.com/VaishnavGhenge/django-silky/blob/master/MIGRATING.md{RESET}
""")


if __name__ == "__main__":
    main()



from django.template import Library
from django.urls import reverse

register = Library()


@register.simple_tag
def sql_detail_url(silk_request, profile, sql_query):
    if profile and silk_request:
        return reverse(
            "silk:request_and_profile_sql_detail",
            args=[silk_request.id, profile.id, sql_query.id],
        )
    elif profile:
        return reverse("silk:profile_sql_detail", args=[profile.id, sql_query.id])
    elif silk_request:
        return reverse("silk:request_sql_detail", args=[silk_request.id, sql_query.id])



import base64
import json
import random
import re
from uuid import uuid4

import sqlparse
from django.conf import settings
from django.core.files.storage import storages
from django.core.files.storage.handler import InvalidStorageError
from django.db import models, router, transaction
from django.db.models import (
    BooleanField,
    CharField,
    DateTimeField,
    FileField,
    FloatField,
    ForeignKey,
    IntegerField,
    ManyToManyField,
    OneToOneField,
    Sum,
    TextField,
)
from django.utils import timezone
from django.utils.safestring import mark_safe

from silk.config import SilkyConfig
from silk.utils.profile_parser import parse_profile

try:
    silk_storage = storages['SILKY_STORAGE']
except InvalidStorageError:
    from django.utils.module_loading import import_string
    storage_class = SilkyConfig().SILKY_STORAGE_CLASS or settings.DEFAULT_FILE_STORAGE
    silk_storage = import_string(storage_class)()


# Seperated out so can use in tests w/o models
def _time_taken(start_time, end_time):
    d = end_time - start_time
    return d.seconds * 1000 + d.microseconds / 1000


def time_taken(self):
    return _time_taken(self.start_time, self.end_time)


class CaseInsensitiveDictionary(dict):
    def __getitem__(self, key):
        return super().__getitem__(key.lower())

    def __setitem__(self, key, value):
        super().__setitem__(key.lower(), value)

    def update(self, other=None, **kwargs):
        for k, v in other.items():
            self[k] = v
        for k, v in kwargs.items():
            self[k] = v

    def __init__(self, d):
        super().__init__()
        for k, v in d.items():
            self[k] = v


class Request(models.Model):
    id = CharField(max_length=36, default=uuid4, primary_key=True)
    path = CharField(max_length=190, db_index=True)
    query_params = TextField(blank=True, default='')
    raw_body = TextField(blank=True, default='')
    body = TextField(blank=True, default='')
    method = CharField(max_length=10)
    start_time = DateTimeField(default=timezone.now, db_index=True)
    view_name = CharField(
        max_length=190, db_index=True, blank=True,
        default='', null=True
    )
    end_time = DateTimeField(null=True, blank=True)
    time_taken = FloatField(blank=True, null=True)  # milliseconds
    encoded_headers = TextField(blank=True, default='')  # stores json
    meta_time = FloatField(null=True, blank=True)
    meta_num_queries = IntegerField(null=True, blank=True)
    meta_time_spent_queries = FloatField(null=True, blank=True)
    pyprofile = TextField(blank=True, default='')
    prof_file = FileField(max_length=300, blank=True, storage=silk_storage)

    # Useful method to create shortened copies of strings without losing start and end context
    # Used to ensure path and view_name don't exceed 190 characters
    def _shorten(self, string):
        return f'{string[:94]}...{string[len(string) - 93:]}'

    @property
    def total_meta_time(self):
        return (self.meta_time or 0) + (self.meta_time_spent_queries or 0)

    @property
    def profile_table(self):
        for n, columns in enumerate(parse_profile(self.pyprofile)):
            location = columns[-1]
            if n and '{' not in location and '<' not in location:
                r = re.compile(r'(?P<src>.*\.py)\:(?P<num>[0-9]+).*')
                m = r.search(location)
                group = m.groupdict()
                src = group['src']
                num = group['num']
                name = 'c%d' % n
                fmt = '<a name={name} href="?pos={n}&file_path={src}&line_num={num}#{name}">{location}</a>'
                rep = fmt.format(**dict(group, **locals()))
                yield columns[:-1] + [mark_safe(rep)]
            else:
                yield columns

    # defined in atomic transaction within SQLQuery save()/delete() as well
    # as in bulk_create of SQLQueryManager
    # TODO: This is probably a bad way to do this, .count() will prob do?
    num_sql_queries = IntegerField(default=0)  # TODO replace with count()

    @property
    def time_spent_on_sql_queries(self):
        """"
        Calculate the total time spent in milliseconds on SQL queries using Django aggregates.
        """
        result = SQLQuery.objects.filter(request=self).aggregate(
            total_time=Sum('time_taken', output_field=FloatField())
        )
        return result['total_time'] or 0.0

    @property
    def headers(self):
        if self.encoded_headers:
            raw = json.loads(self.encoded_headers)
        else:
            raw = {}

        return CaseInsensitiveDictionary(raw)

    @property
    def content_type(self):
        return self.headers.get('content-type', None)

    @classmethod
    def garbage_collect(cls, force=False):
        """ Remove Request/Responses when we are at the SILKY_MAX_RECORDED_REQUESTS limit
        Note that multiple in-flight requests may call this at once causing a
        double collection """
        check_percent = SilkyConfig().SILKY_MAX_RECORDED_REQUESTS_CHECK_PERCENT
        check_percent /= 100.0
        if check_percent < random.random() and not force:
            return
        target_count = SilkyConfig().SILKY_MAX_RECORDED_REQUESTS

        # Since garbage collection is probabilistic, the target count should
        # be lowered to account for requests before the next garbage collection
        if check_percent != 0:
            target_count -= int(1 / check_percent)

        # Make sure we can delete everything if needed by settings
        if target_count <= 0:
            cls.objects.all().delete()
            return

        try:
            time_cutoff = cls.objects.order_by(
                '-start_time'
            ).values_list(
                'start_time',
                flat=True
            )[target_count]
        except IndexError:
            return

        cls.objects.filter(start_time__lte=time_cutoff).delete()

    def save(self, *args, **kwargs):
        # sometimes django requests return the body as 'None'
        if self.raw_body is None:
            self.raw_body = ''

        if self.body is None:
            self.body = ''

        if self.end_time and self.start_time:
            interval = self.end_time - self.start_time
            self.time_taken = interval.total_seconds() * 1000

        # We can't save if either path or view_name exceed 190 characters
        if self.path and len(self.path) > 190:
            self.path = self._shorten(self.path)

        if self.view_name and len(self.view_name) > 190:
            self.view_name = self._shorten(self.view_name)

        super().save(*args, **kwargs)
        Request.garbage_collect(force=False)


class Response(models.Model):
    id = CharField(max_length=36, default=uuid4, primary_key=True)
    request = OneToOneField(
        Request, related_name='response', db_index=True,
        on_delete=models.CASCADE,
    )
    status_code = IntegerField()
    raw_body = TextField(blank=True, default='')
    body = TextField(blank=True, default='')
    encoded_headers = TextField(blank=True, default='')

    @property
    def content_type(self):
        return self.headers.get('content-type', None)

    @property
    def headers(self):
        if self.encoded_headers:
            raw = json.loads(self.encoded_headers)
        else:
            raw = {}
        return CaseInsensitiveDictionary(raw)

    @property
    def raw_body_decoded(self):
        return base64.b64decode(self.raw_body)


# TODO rewrite docstring
class SQLQueryManager(models.Manager):
    def bulk_create(self, *args, **kwargs):
        """ensure that num_sql_queries remains consistent. Bulk create does not call
        the model save() method and hence we must add this logic here too"""
        with transaction.atomic(using=router.db_for_write(SQLQuery)):
            if len(args):
                objs = args[0]
            else:
                objs = kwargs.get('objs')
            for obj in objs:
                obj.prepare_save()

            return super().bulk_create(*args, **kwargs)


class SQLQuery(models.Model):
    query = TextField()
    start_time = DateTimeField(null=True, blank=True, default=timezone.now)
    end_time = DateTimeField(null=True, blank=True)
    time_taken = FloatField(blank=True, null=True)  # milliseconds
    identifier = IntegerField(default=-1)
    request = ForeignKey(
        Request, related_name='queries', null=True,
        blank=True, db_index=True, on_delete=models.CASCADE,
    )
    traceback = TextField()
    analysis = TextField(null=True, blank=True)
    objects = SQLQueryManager()

    # TODO docstring
    @property
    def traceback_ln_only(self):
        return '\n'.join(self.traceback.split('\n')[::2])

    @property
    def formatted_query(self):
        return sqlparse.format(self.query, reindent=True, keyword_case='upper')

    @property
    def num_joins(self):
        parsed_query = sqlparse.parse(self.query)
        count = 0
        for statement in parsed_query:
            count += sum(map(lambda t: t.match(sqlparse.tokens.Keyword, r'\.*join\.*', regex=True), statement.flatten()))
        return count

    @property
    def first_keywords(self):
        parsed_query = sqlparse.parse(self.query)
        keywords = []
        for statement in parsed_query[0].tokens:
            if not statement.is_keyword:
                break
            keywords.append(statement.value)
        return ' '.join(keywords)

    @property
    def tables_involved(self):
        """
        A really another rudimentary way to work out tables involved in a
        query.
        TODO: Can probably parse the SQL using sqlparse etc and pull out table
        info that way?
        """
        components = [x.strip() for x in self.query.split()]
        tables = []

        for idx, component in enumerate(components):
            # TODO: If django uses aliases on column names they will be falsely
            # identified as tables...
            if (
                component.lower() == "from"
                or component.lower() == "join"
                or component.lower() == "as"
                or component.lower() == "update"
            ):
                try:
                    _next = components[idx + 1]
                    if not _next.startswith('('):  # Subquery
                        stripped = _next.strip().strip(',')

                        if stripped:
                            tables.append(stripped)
                except IndexError:  # Reach the end
                    pass
        return tables

    def prepare_save(self):
        if self.end_time and self.start_time:
            interval = self.end_time - self.start_time
            self.time_taken = interval.total_seconds() * 1000

        if not self.pk:
            if self.request:
                self.request.num_sql_queries += 1
                self.request.save(update_fields=['num_sql_queries'])

    def save(self, *args, **kwargs):
        with transaction.atomic(using=router.db_for_write(self)):
            self.prepare_save()
            super().save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        with transaction.atomic(using=router.db_for_write(self)):
            self.request.num_sql_queries -= 1
            self.request.save()
            super().delete(*args, **kwargs)


class BaseProfile(models.Model):
    name = CharField(max_length=300, blank=True, default='')
    start_time = DateTimeField(default=timezone.now)
    end_time = DateTimeField(null=True, blank=True)
    request = ForeignKey(
        Request, null=True, blank=True, db_index=True,
        on_delete=models.CASCADE,
    )
    time_taken = FloatField(blank=True, null=True)  # milliseconds

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if self.end_time and self.start_time:
            interval = self.end_time - self.start_time
            self.time_taken = interval.total_seconds() * 1000
        super().save(*args, **kwargs)


class Profile(BaseProfile):
    file_path = CharField(max_length=300, blank=True, default='')
    line_num = IntegerField(null=True, blank=True)
    end_line_num = IntegerField(null=True, blank=True)
    func_name = CharField(max_length=300, blank=True, default='')
    exception_raised = BooleanField(default=False)
    queries = ManyToManyField(SQLQuery, related_name='profiles', db_index=True)
    dynamic = BooleanField(default=False)

    @property
    def is_function_profile(self):
        return self.func_name is not None

    @property
    def is_context_profile(self):
        return self.func_name is None

    @property
    def time_spent_on_sql_queries(self):
        """
        Calculate the total time spent in milliseconds on SQL queries using Django aggregates.
        """
        result = self.queries.aggregate(
            total_time=Sum('time_taken', output_field=FloatField())
        )
        return result['total_time'] or 0.0



import json

from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.code_generation.curl import curl_cmd
from silk.code_generation.django_test_client import gen
from silk.models import Request, SQLQuery
from silk.utils.n_plus_one import detect_n_plus_one


class RequestView(View):

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, request_id):
        silk_request = Request.objects.get(pk=request_id)
        query_params = None
        if silk_request.query_params:
            query_params = json.loads(silk_request.query_params)

        context = {
            'silk_request': silk_request,
            'query_params': json.dumps(query_params, sort_keys=True, indent=4) if query_params else None,
            'request': request
        }

        if len(silk_request.raw_body) < 20000:  # Don't do this for large request
            body = silk_request.raw_body
            try:
                body = json.loads(body)  # Incase encoded as JSON
            except (ValueError, TypeError):
                pass
            context['curl'] = curl_cmd(url=request.build_absolute_uri(silk_request.path),
                                       method=silk_request.method,
                                       query_params=query_params,
                                       body=body,
                                       content_type=silk_request.content_type)
            context['client'] = gen(path=silk_request.path,
                                    method=silk_request.method,
                                    query_params=query_params,
                                    data=body,
                                    content_type=silk_request.content_type)

        if silk_request.num_sql_queries > 0:
            all_queries = list(SQLQuery.objects.filter(request=silk_request))
            context['n_plus_one'] = detect_n_plus_one(all_queries)

        return render(request, 'silk/request.html', context)



import os
import random
from unittest.mock import patch

from django.conf import settings
from django.test import TestCase

from silk.config import SilkyConfig
from silk.middleware import silky_reverse
from silk.views.sql_detail import SQLDetailView

from .test_lib.mock_suite import MockSuite


class TestViewSQLDetail(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        SilkyConfig().SILKY_AUTHENTICATION = False
        SilkyConfig().SILKY_AUTHORISATION = False

    def test_allowed_file_paths_nothing_specified(self):
        """by default we dont display any source, and it should return correctly"""
        request = MockSuite().mock_request()
        query = MockSuite().mock_sql_queries(request=request, n=1)[0]
        response = self.client.get(silky_reverse('request_sql_detail', kwargs={'sql_id': query.id, 'request_id': request.id}))
        self.assertTrue(response.status_code == 200)

    def test_allowed_file_paths_available_source(self):
        """if we request to view source that exists in the TB all should be fine"""
        request = MockSuite().mock_request()
        query = MockSuite().mock_sql_queries(request=request, n=1)[0]
        tb = query.traceback_ln_only
        _, files = SQLDetailView()._urlify(tb)
        file_path = random.choice(files)
        with open(file_path) as f:
            line_num = random.randint(0, len(f.read().split('\n')))
        response = self.client.get(silky_reverse('request_sql_detail',
                                                 kwargs={'sql_id': query.id, 'request_id': request.id}),
                                   data={
                                       'line_num': line_num,
                                       'file_path': file_path
        })
        self.assertTrue(response.status_code == 200)

    def test_allowed_file_paths_unavailable_source(self):
        """if we request to view source that is not in the traceback we should get a 403"""
        request = MockSuite().mock_request()
        query = MockSuite().mock_sql_queries(request=request, n=1)[0]
        file_path = settings.TEMP_DIR + '/blah'
        with open(file_path, 'w') as f:
            f.write('test')
        response = self.client.get(silky_reverse('request_sql_detail',
                                                 kwargs={'sql_id': query.id, 'request_id': request.id}),
                                   data={
                                       'line_num': 0,
                                       'file_path': file_path
        })
        self.assertTrue(response.status_code == 403)

    def test_virtualenv_not_available_no_highlight(self):
        """if we don't have a virtualenv, there should be no code hightlighted"""
        request = MockSuite().mock_request()
        query = MockSuite().mock_sql_queries(request=request)[0]
        url = silky_reverse('request_sql_detail', kwargs={'sql_id': query.id, 'request_id': request.id})
        with patch.dict(os.environ, {}, clear=True):
            self.assertIsNone(os.environ.get('VIRTUAL_ENV'))
            response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, ' is-third-party')
        self.assertNotContains(response, ' not-third-party')

    def test_virtualenv_hightlight(self):
        """if we have a virtualenv, there should be code hightlighted"""
        request = MockSuite().mock_request()
        query = MockSuite().mock_sql_queries(request=request)[0]
        url = silky_reverse('request_sql_detail', kwargs={'sql_id': query.id, 'request_id': request.id})
        with patch.dict(os.environ, {'VIRTUAL_ENV': '/some/virtualenv/that/doesnt/really/exist'}, clear=True):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, ' not-third-party')



import os
import shutil

from django.db import transaction
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.config import SilkyConfig
from silk.models import Profile, Request, Response, SQLQuery
from silk.utils.data_deletion import delete_model


@method_decorator(transaction.non_atomic_requests, name="dispatch")
class ClearDBView(View):

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, *_, **kwargs):
        return render(request, 'silk/clear_db.html')

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def post(self, request, *_, **kwargs):
        context = {}
        cleared = []

        if 'clear_all' in request.POST:
            delete_model(Profile)
            delete_model(SQLQuery)
            delete_model(Response)
            delete_model(Request)
            cleared = ['Response', 'SQLQuery', 'Profile', 'Request']

            if SilkyConfig().SILKY_DELETE_PROFILES:
                dir = SilkyConfig().SILKY_PYTHON_PROFILER_RESULT_PATH
                for files in os.listdir(dir):
                    path = os.path.join(dir, files)
                    try:
                        shutil.rmtree(path)
                    except OSError:
                        os.remove(path)
                cleared.append('profiler files')
        else:
            if 'clear_requests' in request.POST:
                delete_model(SQLQuery)
                delete_model(Response)
                delete_model(Request)
                cleared += ['Response', 'SQLQuery', 'Request']
            if 'clear_profiling' in request.POST:
                delete_model(Profile)
                cleared.append('Profile')

                if SilkyConfig().SILKY_DELETE_PROFILES:
                    dir = SilkyConfig().SILKY_PYTHON_PROFILER_RESULT_PATH
                    for files in os.listdir(dir):
                        path = os.path.join(dir, files)
                        try:
                            shutil.rmtree(path)
                        except OSError:
                            os.remove(path)
                    cleared.append('profiler files')

        if cleared:
            context['msg'] = 'Cleared data for: {}'.format(', '.join(cleared))

        return render(request, 'silk/clear_db.html', context=context)



import os

from setuptools import find_packages, setup

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-silky',
    use_scm_version=True,
    packages=find_packages(exclude=['project*', 'tests*', 'node_modules*']),
    include_package_data=True,
    license='MIT License',
    description='Silky smooth profiling for the Django Framework — modernized UI fork of django-silk',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/VaishnavGhenge/django-silky',
    author='Michael Ford',
    author_email='mtford@gmail.com',
    maintainer='Vaishnav Ghenge',
    project_urls={
        'Source': 'https://github.com/VaishnavGhenge/django-silky',
        'Bug Tracker': 'https://github.com/VaishnavGhenge/django-silky/issues',
        'Migration Guide': 'https://github.com/VaishnavGhenge/django-silky/blob/master/MIGRATING.md',
        'Upstream': 'https://github.com/jazzband/django-silk',
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 4.2',
        'Framework :: Django :: 5.1',
        'Framework :: Django :: 5.2',
        'Framework :: Django :: 6.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'Django>=4.2',
        'sqlparse',
        'gprof2dot>=2017.09.19',
    ],
    extras_require={
        'formatting': ['autopep8'],
    },
    python_requires='>=3.10',
    setup_requires=['setuptools_scm'],
)



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("django-silky")
except PackageNotFoundError:
    __version__ = version("django-silk")  # fallback for local dev installs



import shlex
from unittest import TestCase

from silk.code_generation.curl import curl_cmd


class TestCodeGenCurl(TestCase):
    def test_post_json(self):
        result = curl_cmd(
            url="https://example.org/alpha/beta",
            method="POST",
            body={"gamma": "delta"},
            content_type="application/json",
        )

        result_words = shlex.split(result)

        self.assertEqual(result_words, [
            'curl', '-X', 'POST',
            '-H', 'content-type: application/json',
            '-d', '{"gamma": "delta"}',
            'https://example.org/alpha/beta'
        ])



#!/usr/bin/env python3
"""
release.py
----------
Automated release script for django-silky.

Usage
-----
  python release.py 1.0.1                  # full release
  python release.py 1.0.1 --test-pypi      # upload to TestPyPI first, then PyPI
  python release.py 1.0.1 --test-pypi-only # upload to TestPyPI only (dry run)
  python release.py 1.0.1 --skip-tests     # skip pytest
  python release.py 1.0.1 --skip-push      # build & upload without pushing tag to remote
  python release.py 1.0.1 --dry-run        # print every step, execute nothing
"""

import argparse
import os
import re
import shutil
import subprocess
import sys
from pathlib import Path

# ── Colours ───────────────────────────────────────────────────────────────────

_COLOUR = sys.stdout.isatty() and not os.environ.get("NO_COLOR")

def _c(code): return code if _COLOUR else ""

RESET  = _c("\033[0m");  BOLD  = _c("\033[1m")
GREEN  = _c("\033[32m"); YELLOW = _c("\033[33m")
RED    = _c("\033[31m"); CYAN  = _c("\033[36m")
DIM    = _c("\033[2m")

def info(msg):    print(f"{CYAN}  ▶  {RESET}{msg}")
def ok(msg):      print(f"{GREEN}  ✔  {RESET}{msg}")
def warn(msg):    print(f"{YELLOW}  ⚠  {RESET}{msg}")
def err(msg):     print(f"{RED}  ✖  {RESET}{msg}", file=sys.stderr)
def heading(msg): print(f"\n{BOLD}{msg}{RESET}")
def dim(msg):     print(f"{DIM}      {msg}{RESET}")
def die(msg, code=1): err(msg); sys.exit(code)

# ── Subprocess helpers ────────────────────────────────────────────────────────

def run(cmd, capture=False, check=True, env=None):
    kwargs = {"check": check, "env": env or os.environ.copy()}
    if capture:
        kwargs["capture_output"] = True
        kwargs["text"] = True
    dim("$ " + " ".join(str(c) for c in cmd))
    return subprocess.run(cmd, **kwargs)

def py(*args, **kwargs):
    return run([sys.executable, *args], **kwargs)

# ── Validation ────────────────────────────────────────────────────────────────

# Matches PEP 440: 1.0.0 / 1.0.0a1 / 1.0.0b2 / 1.0.0rc1 / 1.0.0.post1 / 1.0.0.dev1
_VERSION_RE = re.compile(
    r"^\d+\.\d+\.\d+"
    r"(a\d+|b\d+|rc\d+)?"
    r"(\.post\d+)?"
    r"(\.dev\d+)?$"
)

def validate_version(version: str) -> str:
    if not _VERSION_RE.match(version):
        die(
            f"Invalid version: {version!r}\n"
            "  Expected PEP 440 format, e.g.: 1.0.0 / 1.0.1 / 2.0.0a1 / 1.1.0rc1"
        )
    return version

def tag_exists(version: str) -> bool:
    result = run(["git", "tag", "--list", f"v{version}"], capture=True, check=False)
    return bool(result.stdout.strip())

def working_tree_clean() -> bool:
    result = run(["git", "status", "--porcelain"], capture=True)
    return not result.stdout.strip()

def current_branch() -> str:
    result = run(["git", "rev-parse", "--abbrev-ref", "HEAD"], capture=True)
    return result.stdout.strip()

# ── Steps ─────────────────────────────────────────────────────────────────────

def step_preflight(version: str, args) -> None:
    heading("Step 1 / 6 — Pre-flight checks")

    # Git state
    branch = current_branch()
    ok(f"Branch: {branch}")
    if branch not in ("master", "main"):
        warn(f"Not on master/main (on '{branch}') — are you sure?")

    if not working_tree_clean():
        die(
            "Working tree has uncommitted changes.\n"
            "  Commit or stash them before releasing."
        )
    ok("Working tree is clean")

    # Tag collision
    if tag_exists(version):
        die(
            f"Tag v{version} already exists.\n"
            "  Choose a different version or delete the tag:\n"
            f"    git tag -d v{version} && git push origin :refs/tags/v{version}"
        )
    ok(f"Tag v{version} is available")

    # git must be a system binary
    if not shutil.which("git"):
        die("'git' not found in PATH — install it and retry.")
    ok("Found: git")

    # twine and build are Python packages — check via the current interpreter
    for module, install_name in [("twine", "twine"), ("build", "build")]:
        result = py("-c", f"import {module}", capture=True, check=False)
        if result.returncode != 0:
            die(
                f"'{module}' not installed in current Python environment.\n"
                f"  Run: {sys.executable} -m pip install {install_name}"
            )
        ok(f"Found: {module}")


def step_tests(args) -> None:
    heading("Step 2 / 6 — Run tests")

    if args.skip_tests:
        warn("Skipping tests (--skip-tests).")
        return

    env = os.environ.copy()
    env.setdefault("DB_ENGINE", "sqlite3")

    result = run(
        [sys.executable, "-m", "pytest", "project/tests/", "-q", "--tb=short"],
        check=False, env=env,
    )
    if result.returncode != 0:
        die("Tests failed — aborting release.")
    ok("All tests passed")


def step_tag(version: str, args) -> None:
    heading("Step 3 / 6 — Tag release")

    if args.dry_run:
        info(f"[dry-run] git tag -a v{version} -m 'Release v{version}'")
        return

    run(["git", "tag", "-a", f"v{version}", "-m", f"Release v{version}"])
    ok(f"Created tag: v{version}")

    if args.skip_push:
        warn("Skipping git push (--skip-push).")
    else:
        remote = _get_remote()
        run(["git", "push", remote, "master", "--tags"])
        ok(f"Pushed tag v{version} → {remote}")


def _get_remote() -> str:
    result = run(["git", "remote"], capture=True)
    remotes = result.stdout.strip().splitlines()
    if "origin" in remotes:
        return "origin"
    if remotes:
        return remotes[0]
    die("No git remote configured.")


def step_build(version: str, args) -> Path:
    heading("Step 4 / 6 — Build distributions")

    dist_dir = Path("dist")

    if args.dry_run:
        info("[dry-run] rm -rf dist/ build/ *.egg-info/")
        info("[dry-run] python -m build")
        return dist_dir

    # Clean previous builds
    for d in [dist_dir, Path("build")]:
        if d.exists():
            shutil.rmtree(d)
            info(f"Removed {d}/")
    for egg in Path(".").glob("*.egg-info"):
        shutil.rmtree(egg)
        info(f"Removed {egg}/")

    py("-m", "build")

    # Verify expected files exist
    whl  = list(dist_dir.glob(f"django_silky-{version}-*.whl"))
    sdist = list(dist_dir.glob(f"django_silky-{version}.tar.gz"))

    if not whl or not sdist:
        # setuptools_scm may format version slightly differently; show what's there
        built = list(dist_dir.iterdir())
        ok(f"Built: {[f.name for f in built]}")
    else:
        ok(f"Built: {whl[0].name}")
        ok(f"Built: {sdist[0].name}")

    return dist_dir


def step_check(dist_dir: Path, args) -> None:
    heading("Step 5 / 6 — Check distributions")

    if args.dry_run:
        info("[dry-run] python -m twine check dist/*")
        return

    result = py("-m", "twine", "check", *dist_dir.glob("*"), check=False)
    if result.returncode != 0:
        die("twine check failed — fix the issues above before uploading.")
    ok("twine check passed")


def step_upload(version: str, dist_dir: Path, args) -> None:
    heading("Step 6 / 6 — Upload to PyPI")

    if args.dry_run:
        if args.test_pypi or args.test_pypi_only:
            info("[dry-run] python -m twine upload --repository testpypi dist/*")
        if not args.test_pypi_only:
            info("[dry-run] python -m twine upload dist/*")
        return

    files = list(dist_dir.glob("*"))

    # TestPyPI first
    if args.test_pypi or args.test_pypi_only:
        info("Uploading to TestPyPI …")
        py("-m", "twine", "upload", "--repository", "testpypi", *files)
        ok(f"https://test.pypi.org/project/django-silky/{version}/")

        if args.test_pypi_only:
            warn("--test-pypi-only: stopping here. Run without that flag to publish to PyPI.")
            return

        info("Verify the TestPyPI upload looks correct, then production upload will proceed.")

    # Production PyPI
    info("Uploading to PyPI …")
    py("-m", "twine", "upload", *files)
    ok(f"https://pypi.org/project/django-silky/{version}/")


# ── CLI ───────────────────────────────────────────────────────────────────────

def parse_args():
    parser = argparse.ArgumentParser(
        prog="release.py",
        description="Build and publish a django-silky release to PyPI.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=__doc__,
    )
    parser.add_argument(
        "version",
        help="Version to release, e.g. 1.0.1 (must be PEP 440)",
    )
    parser.add_argument(
        "--test-pypi",
        action="store_true",
        help="Upload to TestPyPI first, then PyPI",
    )
    parser.add_argument(
        "--test-pypi-only",
        action="store_true",
        help="Upload to TestPyPI only (useful for verifying before the real release)",
    )
    parser.add_argument(
        "--skip-tests",
        action="store_true",
        help="Skip running the test suite",
    )
    parser.add_argument(
        "--skip-push",
        action="store_true",
        help="Create the tag locally but do not push to remote",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Print every step without executing anything",
    )
    return parser.parse_args()


# ── Main ──────────────────────────────────────────────────────────────────────

def main():
    args = parse_args()
    version = validate_version(args.version)

    print(f"""
{BOLD}┌────────────────────────────────────────┐
│  django-silky release script  v{version:<8} │
└────────────────────────────────────────┘{RESET}""")

    if args.dry_run:
        warn("DRY-RUN mode — no changes will be made.\n")

    step_preflight(version, args)
    step_tests(args)
    step_tag(version, args)
    dist_dir = step_build(version, args)
    step_check(dist_dir, args)
    step_upload(version, dist_dir, args)

    print(f"""
{GREEN}{BOLD}  Released django-silky {version}!{RESET}

  PyPI:    https://pypi.org/project/django-silky/{version}/
  GitHub:  https://github.com/VaishnavGhenge/django-silky/releases/tag/v{version}

  Install: pip install django-silky=={version}
""")


if __name__ == "__main__":
    main()



from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import View

from silk.auth import login_possibly_required, permissions_possibly_required
from silk.models import Profile, Request, SQLQuery
from silk.utils.n_plus_one import detect_n_plus_one
from silk.utils.pagination import _page

__author__ = 'mtford'


class SQLView(View):
    page_sizes = [5, 10, 25, 100, 200, 500, 1000]
    default_page_size = 200

    @method_decorator(login_possibly_required)
    @method_decorator(permissions_possibly_required)
    def get(self, request, *_, **kwargs):
        request_id = kwargs.get('request_id')
        profile_id = kwargs.get('profile_id')
        per_page_param = request.GET.get('per_page')
        if per_page_param is not None:
            try:
                per_page = int(per_page_param)
            except (TypeError, ValueError):
                per_page = self.default_page_size
            request.session['silk_sql_per_page'] = per_page
        else:
            try:
                per_page = int(request.session.get('silk_sql_per_page', self.default_page_size))
            except (TypeError, ValueError):
                per_page = self.default_page_size
        context = {
            'request': request,
            'options_page_size': self.page_sizes,
            'per_page': per_page,
        }
        if request_id:
            silk_request = Request.objects.get(id=request_id)
            all_queries = list(SQLQuery.objects.filter(request=silk_request).order_by('-start_time'))
            for q in all_queries:
                q.start_time_relative = q.start_time - silk_request.start_time
            n_plus_one = detect_n_plus_one(all_queries)
            page = _page(request, all_queries, per_page)
            context['silk_request'] = silk_request
            context['n_plus_one'] = n_plus_one
            context['n_plus_one_ids'] = n_plus_one.flagged_query_ids
        if profile_id:
            p = Profile.objects.get(id=profile_id)
            page = _page(request, p.queries.order_by('-start_time').all(), per_page)
            context['profile'] = p
        if not (request_id or profile_id):
            raise KeyError('no profile_id or request_id')
        # noinspection PyUnboundLocalVariable
        context['items'] = page
        return render(request, 'silk/sql.html', context)



import cProfile
import os.path
import sys

from django.test import TestCase
from tests.util import DictStorage

from silk.collector import DataCollector
from silk.config import SilkyConfig

from .factories import RequestMinFactory


class TestCollector(TestCase):
    def test_singleton(self):
        a = DataCollector()
        b = DataCollector()
        c = DataCollector()
        self.assertTrue(a == b == c)

    def test_query_registration(self):
        mock_query = {}
        DataCollector().register_query(mock_query)
        self.assertIn(mock_query, list(DataCollector().queries.values()))

    def test_clear(self):
        self.test_query_registration()
        DataCollector().clear()
        self.assertFalse(DataCollector().queries)

    def test_finalise(self):
        request = RequestMinFactory()
        DataCollector().configure(request)
        with self.subTest("Default file-based storage"):
            DataCollector().finalise()
            file = DataCollector().request.prof_file
            self.assertIsNotNone(file)
            with file.storage.open(file.name) as f:
                content = f.read()
                self.assertTrue(content)

        # Some storages, such as S3Boto3Storage, don't support local file system path.
        # Simulate this behaviour using DictStorage.
        with self.subTest("Pathless storage"):
            request.prof_file.storage = DictStorage()
            DataCollector().finalise()
            file = DataCollector().request.prof_file
            self.assertIsNotNone(file)
            with file.storage.open(file.name) as f:
                content = f.read()
                self.assertTrue(content)
                self.assertGreater(len(content), 0)

    def test_configure_exception(self):
        other_profiler = cProfile.Profile()
        other_profiler.enable()
        collector = DataCollector()
        collector.configure()
        other_profiler.disable()
        if sys.version_info >= (3, 12):
            self.assertEqual(collector.local.pythonprofiler, None)
        else:
            self.assertIsNotNone(collector.local.pythonprofiler)
            collector.stop_python_profiler()

    def test_profile_file_name_with_disabled_extended_file_name(self):
        SilkyConfig().SILKY_PYTHON_PROFILER_EXTENDED_FILE_NAME = False
        request_path = 'normal/uri/'
        resulting_prefix = self._get_prof_file_name(request_path)
        self.assertEqual(resulting_prefix, '')

    def test_profile_file_name_with_enabled_extended_file_name(self):

        SilkyConfig().SILKY_PYTHON_PROFILER_EXTENDED_FILE_NAME = True
        request_path = 'normal/uri/'
        resulting_prefix = self._get_prof_file_name(request_path)
        self.assertEqual(resulting_prefix, 'normal_uri_')

    def test_profile_file_name_with_path_traversal_and_special_char(self):
        SilkyConfig().SILKY_PYTHON_PROFILER_EXTENDED_FILE_NAME = True
        request_path = 'spÉciàl/.././大/uri/@É/'
        resulting_prefix = self._get_prof_file_name(request_path)
        self.assertEqual(resulting_prefix, 'special_uri_e_')

    def test_profile_file_name_with_long_path(self):
        SilkyConfig().SILKY_PYTHON_PROFILER_EXTENDED_FILE_NAME = True
        request_path = 'long/path/' + 'a' * 100
        resulting_prefix = self._get_prof_file_name(request_path)
        # the path is limited to 50 char plus the last `_`
        self.assertEqual(len(resulting_prefix), 51)

    @classmethod
    def _get_prof_file_name(cls, request_path: str) -> str:
        request = RequestMinFactory()
        request.path = request_path
        DataCollector().configure(request)
        DataCollector().finalise()
        file_path = DataCollector().request.prof_file.name
        filename = os.path.basename(file_path)
        return filename.replace(f"{request.id}.prof", "")
